@@ -18,7 +18,7 @@ static_assert(false, "standalone or boost asio must be included to use glaze/ext
 #include <coroutine>
 #include <iostream>
 
-#include "glaze/rpc/repe/registry.hpp"
+#include "glaze/rpc/registry.hpp"
 #include "glaze/util/memory_pool.hpp"
 
 namespace glz
@@ -43,10 +43,7 @@ namespace glz
          }
          const auto error_size = 4 + error_message.size(); // 4 bytes for message length
          msg.header.body_length = error_size;
-         msg.body.resize(error_size);
-         const uint32_t n = uint32_t(error_message.size());
-         std::memcpy(msg.body.data(), &n, 4);
-         std::memcpy(msg.body.data() + 4, error_message.data(), n);
+         msg.body = error_message;
       }
 
       template <class ErrorMessage>
@@ -62,9 +59,7 @@ namespace glz
          if (bool(msg.error())) {
             const auto ec = msg.header.ec;
             if (msg.header.body_length >= 4) {
-               uint32_t error_length{};
-               std::memcpy(&error_length, msg.body.data(), 4);
-               const std::string_view error_message{msg.body.data() + 4, error_length};
+               const std::string_view error_message = msg.body;
                std::string ret = "REPE error: ";
                ret.append(format_error(ec));
                ret.append(" | ");
@@ -80,15 +75,13 @@ namespace glz
 
       // Decodes a repe::message into a structure
       // Returns a std::string with a formatted error on error
-      template <opts Opts = opts{}, class T>
+      template <auto Opts = opts{}, class T>
       inline std::optional<std::string> decode_message(T&& value, message& msg)
       {
          if (bool(msg.header.ec)) {
             const auto ec = msg.header.ec;
             if (msg.header.body_length >= 4) {
-               uint32_t error_length{};
-               std::memcpy(&error_length, msg.body.data(), 4);
-               const std::string_view error_message{msg.body.data() + 4, error_length};
+               const std::string_view error_message = msg.body;
                std::string ret = "REPE error: ";
                ret.append(format_error(ec));
                ret.append(" | ");
@@ -111,11 +104,15 @@ namespace glz
    }
 
 #if defined(GLZ_USING_BOOST_ASIO)
-   namespace asio = boost::asio;
+   namespace asio
+   {
+      using namespace boost::asio;
+      using error_code = boost::system::error_code;
+   }
 #endif
    inline void send_buffer(asio::ip::tcp::socket& socket, repe::message& msg)
    {
-      if (msg.header.length == repe::no_length_provided) {
+      if (msg.header.length == 0) {
          encode_error(error_code::invalid_header, msg);
          return;
       }
@@ -139,16 +136,8 @@ namespace glz
          encode_error(error_code::connection_failure, msg, e.message());
          return;
       }
-      if (msg.header.query_length == repe::no_length_provided) {
-         encode_error(error_code::invalid_query, msg);
-         return;
-      }
       msg.query.resize(msg.header.query_length);
       asio::read(socket, asio::buffer(msg.query), asio::transfer_exactly(msg.header.query_length));
-      if (msg.header.body_length == repe::no_length_provided) {
-         encode_error(error_code::invalid_body, msg);
-         return;
-      }
       msg.body.resize(msg.header.body_length);
       asio::read(socket, asio::buffer(msg.body), asio::transfer_exactly(msg.header.body_length), e);
       if (e) {
@@ -159,7 +148,7 @@ namespace glz
 
    inline asio::awaitable<void> co_send_buffer(asio::ip::tcp::socket& socket, const repe::message& msg)
    {
-      if (msg.header.length == repe::no_length_provided) {
+      if (msg.header.length == 0) {
          throw std::runtime_error("No length provided in REPE header");
       }
       if (msg.header.query_length != msg.query.size()) {
@@ -187,17 +176,11 @@ namespace glz
                                 asio::transfer_exactly(sizeof(msg.header)), asio::use_awaitable);
 
       // Validate and read the query
-      if (msg.header.query_length == repe::no_length_provided) {
-         throw std::runtime_error("No query_length provided in REPE header");
-      }
       msg.query.resize(msg.header.query_length);
       co_await asio::async_read(socket, asio::buffer(msg.query), asio::transfer_exactly(msg.header.query_length),
                                 asio::use_awaitable);
 
       // Validate and read the body
-      if (msg.header.body_length == repe::no_length_provided) {
-         throw std::runtime_error("No body_length provided in REPE header");
-      }
       msg.body.resize(msg.header.body_length);
       co_await asio::async_read(socket, asio::buffer(msg.body), asio::transfer_exactly(msg.header.body_length),
                                 asio::use_awaitable);
@@ -220,8 +203,13 @@ namespace glz
       {
          std::unique_lock lock{mtx};
 
+         if (not ctx) {
+            // TODO: make this error into an error code
+            throw std::runtime_error("asio::io_context is null");
+         }
+
          // reset all socket pointers if a connection failed
-         if (not *is_connected) {
+         if (not*is_connected) {
             for (auto& socket : sockets) {
                socket.reset();
             }
@@ -304,7 +292,7 @@ namespace glz
       ~unique_socket() { pool->free(index); }
    };
 
-   template <opts Opts = opts{}>
+   template <auto Opts = opts{}>
    struct asio_client
    {
       std::string host{"localhost"}; // host name
@@ -318,7 +306,7 @@ namespace glz
       };
 
       std::shared_ptr<asio::io_context> ctx{};
-      std::shared_ptr<glz::socket_pool> socket_pool = std::make_shared<glz::socket_pool>();
+      std::shared_ptr<glz::socket_pool> socket_pool{};
       std::shared_ptr<glz::memory_pool<repe::message>> message_pool =
          std::make_shared<glz::memory_pool<repe::message>>();
 
@@ -329,11 +317,19 @@ namespace glz
 
       [[nodiscard]] error_code init()
       {
-         ctx = std::make_shared<asio::io_context>(concurrency);
-         socket_pool->ctx = ctx;
-         socket_pool->host = host;
-         socket_pool->service = service;
-         is_connected = socket_pool->is_connected;
+         *is_connected = false;
+         // create a new socket_pool if we are initilaizing, this is needed because the sockets hold references to the
+         // io_context which is being recreated with init()
+         socket_pool = std::make_shared<glz::socket_pool>();
+
+         {
+            std::unique_lock lock{socket_pool->mtx}; // lock the socket_pool when setting up
+            ctx = std::make_shared<asio::io_context>(concurrency);
+            socket_pool->ctx = ctx;
+            socket_pool->host = host;
+            socket_pool->service = service;
+            is_connected = socket_pool->is_connected;
+         }
 
          unique_socket socket{socket_pool.get()};
          if (socket) {
@@ -348,6 +344,11 @@ namespace glz
       void call(Header&& header, repe::message& response, Params&&... params)
       {
          auto request = message_pool->borrow();
+         request->body.clear();
+         if (not connected()) {
+            encode_error(request->error(), response, "call failure: NOT CONNECTED");
+            return;
+         }
          repe::request<Opts>(std::move(header), *request, std::forward<Params>(params)...);
          if (bool(request->error())) {
             encode_error(request->error(), response, "bad request");
@@ -365,26 +366,168 @@ namespace glz
          send_buffer(*socket, *request);
 
          if (bool(request->error())) {
-            socket.ptr.reset();
-            (*is_connected) = false;
             return;
          }
 
          if (not header.notify) {
             receive_buffer(*socket, response);
             if (bool(response.error())) {
-               socket.ptr.reset();
-               (*is_connected) = false;
                return;
             }
          }
       }
+
+      // The `call` method above is designed to be generic and fully-featured.
+      // The `call` method handles invoking RPC functions and sending data to query endpoints
+      // as well as receiving data back.
+      // The API is designed to reduce copies and gives the user control of how the response is handled.
+      // This is important for chaining calls and efficiently handling memory.
+      // ---
+      // The `send` and `receive` methods provide simpler APIs that throw for cleaner code
+      // and allocate the response in the function call.
+
+      // Send parameters to a target query function or value
+      template <class... Params>
+      void set(const std::string_view query, Params&&... params)
+      {
+         if (not connected()) {
+            throw std::runtime_error("asio_client: NOT CONNECTED");
+         }
+
+         repe::message response{};
+         auto request = message_pool->borrow();
+         request->body.clear();
+         repe::request<Opts>(repe::user_header{.query = query}, *request, std::forward<Params>(params)...);
+         if (bool(request->error())) {
+            throw std::runtime_error("bad request");
+         }
+
+         unique_socket socket{socket_pool.get()};
+         if (not socket) {
+            socket.ptr.reset();
+            (*is_connected) = false;
+            throw std::runtime_error("socket failure");
+         }
+
+         send_buffer(*socket, *request);
+
+         if (bool(request->error())) {
+            throw std::runtime_error(glz::format_error(request->error()));
+         }
+
+         receive_buffer(*socket, response);
+         if (bool(response.error())) {
+            std::string error_message = glz::format_error(response.error());
+            if (response.body.size()) {
+               error_message.append(": ");
+               error_message.append(response.body);
+            }
+            throw std::runtime_error(error_message);
+         }
+      }
+
+      template <class Output>
+      void get(const std::string_view query, Output& output)
+      {
+         if (not connected()) {
+            throw std::runtime_error("asio_client: NOT CONNECTED");
+         }
+
+         repe::message response{};
+         auto request = message_pool->borrow();
+         request->body.clear();
+         repe::request<Opts>(repe::user_header{.query = query}, *request);
+         if (bool(request->error())) {
+            throw std::runtime_error("bad request");
+         }
+
+         unique_socket socket{socket_pool.get()};
+         if (not socket) {
+            socket.ptr.reset();
+            (*is_connected) = false;
+            throw std::runtime_error("socket failure");
+         }
+
+         send_buffer(*socket, *request);
+
+         if (bool(request->error())) {
+            throw std::runtime_error(glz::format_error(request->error()));
+         }
+
+         receive_buffer(*socket, response);
+         if (bool(response.error())) {
+            std::string error_message = glz::format_error(response.error());
+            if (response.body.size()) {
+               error_message.append(": ");
+               error_message.append(response.body);
+            }
+            throw std::runtime_error(error_message);
+         }
+
+         auto ec = read<Opts>(output, response.body);
+         if (bool(ec)) {
+            throw std::runtime_error(glz::format_error(ec, response.body));
+         }
+      }
+
+      // Allocating version of `get`
+      template <class Output>
+      auto get(const std::string_view query)
+      {
+         Output out{};
+         get(query, out);
+         return out;
+      }
+
+      template <class Input, class Output>
+      void inout(const std::string_view query, Input&& input, Output& output)
+      {
+         if (not connected()) {
+            throw std::runtime_error("asio_client: NOT CONNECTED");
+         }
+
+         repe::message response{};
+         auto request = message_pool->borrow();
+         request->body.clear();
+         repe::request<Opts>(repe::user_header{.query = query}, *request, std::forward<Input>(input));
+         if (bool(request->error())) {
+            throw std::runtime_error("bad request");
+         }
+
+         unique_socket socket{socket_pool.get()};
+         if (not socket) {
+            socket.ptr.reset();
+            (*is_connected) = false;
+            throw std::runtime_error("socket failure");
+         }
+
+         send_buffer(*socket, *request);
+
+         if (bool(request->error())) {
+            throw std::runtime_error(glz::format_error(request->error()));
+         }
+
+         receive_buffer(*socket, response);
+         if (bool(response.error())) {
+            std::string error_message = glz::format_error(response.error());
+            if (response.body.size()) {
+               error_message.append(": ");
+               error_message.append(response.body);
+            }
+            throw std::runtime_error(error_message);
+         }
+
+         auto ec = read<Opts>(output, response.body);
+         if (bool(ec)) {
+            throw std::runtime_error(glz::format_error(ec, response.body));
+         }
+      }
    };
 
-   template <opts Opts = opts{}>
+   template <auto Opts = opts{}>
    struct asio_server
    {
-      uint16_t port{};
+      uint16_t port{}; // 0 will select a random free port
       uint32_t concurrency{1}; // How many threads to use (a call to .run() is inclusive on the main thread)
 
       // Register a callback that takes a string error message on server/registry errors.
@@ -405,12 +548,12 @@ namespace glz
       std::shared_ptr<asio::signal_set> signals{};
       std::shared_ptr<std::vector<std::thread>> threads{};
 
-      repe::registry<Opts> registry{};
+      glz::registry<Opts> registry{};
 
       void clear_registry() { registry.clear(); }
 
-      template <const std::string_view& Root = repe::detail::empty_path, class T>
-         requires(glz::detail::glaze_object_t<T> || glz::detail::reflectable<T>)
+      template <const std::string_view& Root = detail::empty_path, class T>
+         requires(glaze_object_t<T> || reflectable<T>)
       void on(T& value)
       {
          registry.template on<Root>(value);
@@ -439,8 +582,13 @@ namespace glz
          // Setup signal handling to stop the server
          signals->async_wait([&](auto, auto) { ctx->stop(); });
 
+         // Create the acceptor synchronously so we know the actual port if set to 0 (select random free)
+         auto executor = ctx->get_executor();
+         asio::ip::tcp::acceptor acceptor(executor, {asio::ip::tcp::v6(), port});
+         port = acceptor.local_endpoint().port();
+
          // Start the listener coroutine
-         asio::co_spawn(*ctx, listener(), asio::detached);
+         asio::co_spawn(*ctx, listener(std::move(acceptor)), asio::detached);
 
          // Run the io_context in multiple threads for concurrency
          threads = std::shared_ptr<std::vector<std::thread>>(new std::vector<std::thread>{}, [](auto* ptr) {
@@ -478,14 +626,17 @@ namespace glz
       asio::awaitable<void> run_instance(asio::ip::tcp::socket socket)
       {
          socket.set_option(asio::ip::tcp::no_delay(true));
+
+         // Allocate once and reuse memory
          repe::message request{};
          repe::message response{};
 
          try {
             while (true) {
                co_await co_receive_buffer(socket, request);
+               response.header.ec = {}; // clear error code, as we use this field to determine if a new error occured
                registry.call(request, response);
-               if (not request.header.notify()) {
+               if (not request.header.notify) {
                   co_await co_send_buffer(socket, response);
                }
             }
@@ -500,10 +651,10 @@ namespace glz
          }
       }
 
-      asio::awaitable<void> listener()
+      asio::awaitable<void> listener(asio::ip::tcp::acceptor acceptor)
       {
          auto executor = co_await asio::this_coro::executor;
-         asio::ip::tcp::acceptor acceptor(executor, {asio::ip::tcp::v6(), port});
+
          while (true) {
             auto socket = co_await acceptor.async_accept(asio::use_awaitable);
             asio::co_spawn(executor, run_instance(std::move(socket)), asio::detached);
